#ifndef NDSMENU_H_
#define NDSMENU_H_
#include "nds.h"

#define MAX_ENTRIES  12
#define MAX_LENGTH  100
#define MAX_SEL_ITEMS  12

class Menu;

class menuItem{
	public:
		Menu * parent;
//		virtual subItem (char * text)=0;
		virtual char* getText()=0;
		virtual char* getHelpText()=0;
		virtual int execute(const uint16 keys)=0;
		virtual void cancel()=0;
		virtual bool hasSelection()=0;
		virtual bool isActive()=0;
		virtual ~menuItem(){}
};

class Menu{
	protected:
		menuItem *subItems[MAX_ENTRIES];
		unsigned int posx, posy;
		unsigned short entryCount;
		char* name;
		Menu * parent;
		
		virtual void drawItem (int item, bool selected);
		void redraw (int selItem);
		virtual int buttonAHandler (int selItem);
	public:
		Menu(char * _name);
		void addMenuItem (menuItem * item);
		int display();
		void setPosition(unsigned int px, unsigned int py){posx = px; posy = py;}
		virtual ~Menu(){}
};

class subMenuItem: public menuItem{
	private:
		char * name;
		char * help;
		Menu * sMenu;
	public:
		bool active;
		bool isActive() {return active;}
		subMenuItem(char * _name, char * _help, Menu * _sMenu):name(_name), help(_help), sMenu(_sMenu), active(true){}
		void setSubMenu(Menu * _sMenu);
		char* getText(){return name;}
		char* getHelpText(){ return help;}
		int execute(const uint16 keys);
		void cancel(){}
		bool hasSelection(){return false;}
};

class selectionMenu: public Menu{
	private:
		menuItem * choosen;
		virtual void drawItem (int item, bool selected);
		virtual int buttonAHandler (int selItem);
	public:
		selectionMenu(char * _name):Menu(_name){}
		menuItem* getChoosen(){ return choosen;}
		int getChoosenID(){
			int t=0;
			while (subItems[t] != choosen) t++;
			return t;
		}
		void setChoosen(menuItem * choose){ choosen = choose;}
		void setChoosenID(int id){ choosen = subItems[id];}

};

/**
 * This one is usefull for e.g. selectionMenus
 */
class SimpleMenuItem:public menuItem{
	protected:
		char * name, *help;
	public:
		bool active;
		int retValue;
		bool isActive() {return active;}
		SimpleMenuItem(char * _name, char * _help):name(_name),help(_help), active(true), retValue(0){}
		virtual char* getText(){ return name;}
		char* getHelpText(){ return help;}
		virtual bool hasSelection(){ return false;}
		virtual int execute(const uint16 keys){ return retValue;}
		virtual void cancel(){}
};

class SimpleSelectionMenuItem:public SimpleMenuItem{
	private:
		char* selItems[MAX_SEL_ITEMS];
		unsigned short entryCount;
		unsigned short selectedItem;
		char buf[256];
	public: 
		SimpleSelectionMenuItem(char * _name, char * _help):SimpleMenuItem(_name, _help){}
		char* getText();
		bool hasSelection(){ return true;}
		int execute(const uint16 keys);
		void addSelectionItem(char * selItem);
		unsigned short getSelection(){return selectedItem;}
		void setSelection(unsigned short selection){ selectedItem = selection;}
};
#endif /*NDSMENU_H_*/
