
#include "nds.h"
#include "pieces.h"
#include "move.h"
#include "board.h"
#include "piece_Moves.h"
#include "chessgfx.h"
#include "fonts.h"
#include "put_raw.h"
#include "take_raw.h"
#include "san.h"

#define MAX_SPRITES 64
//a global copy of sprite attribute memory
static SpriteEntry sprites[MAX_SPRITES];
//rotation attributes overlap so assign then to the same location
pSpriteRotation spriteRotations = (pSpriteRotation)sprites;
int spriteCount = 0;

int lastmove;

TransferSoundData put = {
		put_raw,		/* Sample address */
		put_raw_size,	/* Sample length */
		44100,				/* Sample rate */
		100,				/* Volume */
		64,					/* panning */
		2 					/* format */
	};
TransferSoundData take = {
		take_raw,		/* Sample address */
		take_raw_size,	/* Sample length */
		44100,				/* Sample rate */
		30,				/* Volume */
		64,					/* panning */
		2 					/* format */
	};


void hideMove ();


int getLastMove(){
	return lastmove;
}

void resetLastMove(){
	lastmove = MoveNone;
}

void initSprites(void){
	for(int i = 0; i < MAX_SPRITES; i++){
		sprites[i].attribute[0] = ATTR0_DISABLED | ATTR0_COLOR_256;
		sprites[i].attribute[1] = ATTR1_SIZE_16;
		sprites[i].attribute[2] = 0;
		sprites[i].attribute[3] = 0;
	}

 	for (int x=0; x<6;x++){
		for(uint16 i=0; i<64; i++){
			SPRITE_GFX[128*x*2 + i] = pieces_gfx[8*96 + i+x*64];
			SPRITE_GFX[128*x*2 + 64 + i] = pieces_gfx[12*96+i+x*64];
			SPRITE_GFX[128*(x*2+1) + i] = pieces_gfx[i+x*64];
			SPRITE_GFX[128*(x*2+1) + 64 + i] = pieces_gfx[4*96+i+x*64];
		}
	}
	for(uint16 i=0; i<255; i++) SPRITE_PALETTE[i] = pieces_pal[i];
	spriteCount = 0;
}

/**
 * Copy our sprites to object attribute memory
 */
void updateOAM(void){
	DC_FlushAll();
    dmaCopy(sprites, OAM, MAX_SPRITES * sizeof(SpriteEntry));
}

const uint8 BOARD_TOP = 14;
const uint8 BOARD_LEFT = 51;
const uint8 SQUARE_SIZE = 20;

struct Point{
	int x;
	int y;
};

/**
 * Returns the (x,y)-Coordinates of given chess square
 */
Point calcBoardxy(const int square64, const bool boardRotated){
	Point point;
	const int p = square64 % 8;
	const int pp = square64 / 8;
	if (boardRotated){ 
		point.x = (7 - p)*SQUARE_SIZE + BOARD_LEFT;
		point.y = pp*SQUARE_SIZE + BOARD_TOP;
	}
	else {
		point.x = p*SQUARE_SIZE + BOARD_LEFT;
		point.y = (7 - pp)*SQUARE_SIZE + BOARD_TOP;
	}
	return point;
}

int getBoardposi (const int x, const int y, bool black){
	if ((x-BOARD_LEFT) < 0 || (x-BOARD_LEFT) > 8*SQUARE_SIZE) return -1;
	else if ((y-BOARD_TOP) < 0 || (y-BOARD_TOP) > 8*SQUARE_SIZE) return -1;
	if (black) return 63-((x - BOARD_LEFT)/SQUARE_SIZE + (7-(y-BOARD_TOP)/SQUARE_SIZE)*8);
	return (x - BOARD_LEFT)/SQUARE_SIZE + (7-(y-BOARD_TOP)/SQUARE_SIZE)*8;
}

void clearSprites(){
	for(int i = 0; i < spriteCount; i++){
		sprites[i].attribute[0] = ATTR0_DISABLED | ATTR0_COLOR_256;
	}
	spriteCount = 0;
}

/**
 * Creates a new chess piece and sets it into position
 */
void createSprite(const int piece, const int spc, const int x, const int y) {
	sprites[spc].attribute[0] = ATTR0_COLOR_256 | ATTR0_ROTSCALE | y;
	sprites[spc].attribute[1] = ATTR1_SIZE_16 | x;
	spriteRotations[spc].hdx=256;
	spriteRotations[spc].hdy=0;
	spriteRotations[spc].vdx=0;
	spriteRotations[spc].vdy=256;
	if (PIECE_IS_PAWN(piece))
		sprites[spc].attribute[2] = 8*(WhitePawn12 + PIECE_COLOUR(piece));
	else if (PIECE_IS_KNIGHT(piece))
		sprites[spc].attribute[2] = 8*(WhiteKnight12 + PIECE_COLOUR(piece));
	else if (PIECE_IS_BISHOP(piece))
		sprites[spc].attribute[2] = 8*(WhiteBishop12 + PIECE_COLOUR(piece));
	else if (PIECE_IS_QUEEN(piece))
		sprites[spc].attribute[2] = 8*(WhiteQueen12 + PIECE_COLOUR(piece));
	else if (PIECE_IS_KING(piece))
		sprites[spc].attribute[2] = 8*(WhiteKing12 + PIECE_COLOUR(piece));
	else if (PIECE_IS_ROOK(piece))
		sprites[spc].attribute[2] = 8*(WhiteRook12 + PIECE_COLOUR(piece));
}

/**
 * Draws the sprites on the screen
 */
void fill_board(const board_t * board, const bool black){
	clearSprites();
	for (int i=0; i<64; i++){
		const int piece = board->square[SQUARE_FROM_64(i)];
		//if (black) piece = board->square[SQUARE_FROM_64(63-i)];
		if (PIECE_TYPE(piece)>0){
			const Point p = calcBoardxy(i, black);
			createSprite(piece, spriteCount++, p.x + 2, p.y + 2);
		}
	}
	swiWaitForVBlank();
	updateOAM();
}

/**
 * Draws the captured pieces at the side
 */
void drawCapturedPieces(const int* pieces, const int count, const bool isBoardRotated) {
	//Captured pieces
	int cwhite=0;
	for (int i=0; i<count; i++) {
		const int piece = pieces[i];
		const bool isWhite = !PIECE_COLOUR(piece);
		int y = BOARD_TOP;
		int x = BOARD_LEFT;
		if ((isWhite && !isBoardRotated) || (!isWhite && isBoardRotated)) {
			y += 7*SQUARE_SIZE;
			y -= (cwhite/2)*((2*SQUARE_SIZE)/3 + 2);
			x += 8*SQUARE_SIZE + SQUARE_SIZE/2;
			x += (cwhite % 2)*(2*SQUARE_SIZE)/3;
			cwhite++;
		} 
		else {
			const int p = i-cwhite;
			y += (p/2)*((2*SQUARE_SIZE)/3 + 2);
			x -= SQUARE_SIZE + SQUARE_SIZE/2;
			x -= (p % 2)*(2*SQUARE_SIZE)/3;			
		}
		createSprite(piece, spriteCount++, x + 2, y + 2);
	}
	swiWaitForVBlank();
	updateOAM();
}

// reading pixel position:

static int touchx;
static int touchy;

/**
 * Variant for usage with new ndslib
 */
static void updateTouch(){
	touchx = IPC->touchXpx;
	touchy = IPC->touchYpx;
}

int getDragSprite (const int x, const int y){
	//do not allow dragging of sprites outside the board
	if (x < BOARD_LEFT || x > BOARD_LEFT + 8*SQUARE_SIZE || y < BOARD_TOP || y > BOARD_TOP + 8*SQUARE_SIZE)
		return -1;   
	for(int i = 0; i < spriteCount; i++){
		if ((sprites[i].attribute[0] & ATTR0_ROTSCALE) && (sprites[i].attribute[0] & 255) <= y && 
			((sprites[i].attribute[0] & 255) + 16) >= y &&
			(sprites[i].attribute[1] & 511) <= x && ((sprites[i].attribute[1] & 511) + 16) >= x)
			return i;
	}
	return -1;
}

/**
 * Move sprite s to position x,y
 */ 
void updateSprite(const int s, const uint16 x, const uint16 y){
	sprites[s].attribute[0] = ATTR0_COLOR_256 | ATTR0_ROTSCALE | y;
	sprites[s].attribute[1] = ATTR1_SIZE_16 | x;
	
	for (unsigned int i=0; i< sizeof(SpriteEntry) / 4; i++)
		((u32*)OAM)[s * sizeof(SpriteEntry)/4 + i] = ((u32*)sprites)[s * sizeof(SpriteEntry)/4 + i];
}


#define PEN_HIT (~pen_curr & pen_prev)
#define PEN_RELEASED (pen_curr & ~pen_prev)

/**
 * Generates a list and shows all legal moves 
 * from the selected position
 */ 
bool showMoves(const int square64, board_t * board, const bool isBoardRotated){
	list_t list;
	list.size = 0;
	piece_Moves(&list, board, SQUARE_FROM_64(square64));
	if (list.size == 0) return false;
	swiWaitForVBlank();
	for (int i=0;i<list.size;i++){
		highlightField(SQUARE_TO_64(MOVE_TO(list.move[i])), false, isBoardRotated);
	}
	return true;
}

/**
 * Returns true if the human player has made a legal move.
 * You can get this move by using getLastMove()
 */
bool waitT(board_t * board, const bool isBoardRotated){
	u16 x, y ;
	static u16 pen_prev = IPC_PEN_DOWN ;
	static u16 pen_curr = IPC_PEN_DOWN ;
	int sprite=-1;
	int oldPosi = -1;

	for (;;) {
		pen_prev = pen_curr ;
		pen_curr = (IPC->buttons & IPC_PEN_DOWN);
    	if (PEN_HIT) {
    		updateTouch();
			x = touchx;
      		y = touchy;
      		sprite = getDragSprite(x,y);
      		if (sprite >=0){
      			oldPosi = getBoardposi(x, y, isBoardRotated);
      			showMoves(oldPosi, board, isBoardRotated);
      			playSound(&take);
      		}
			else return false;
		}
		if (PEN_RELEASED && sprite >= 0){
			//int posi = getBoardposi(x, y);
			hideMove();
			int posi = getBoardposi ((sprites[sprite].attribute[1] + SQUARE_SIZE/2) & 511, 
				(sprites[sprite].attribute[0] + SQUARE_SIZE/2) & 255, isBoardRotated);
			if (posi<0 || posi > 64) posi = oldPosi;
			else if (posi != oldPosi){
				int m = check_move(board, oldPosi, posi);
				if (m == MoveNone) posi = oldPosi;
				else {
					lastmove = m;
					//Play sound
					playSound(&put);
				}
			}
			const Point p = calcBoardxy(posi, isBoardRotated);
			updateSprite (sprite, p.x + 2, p.y + 2);
			updateOAM();
			sprite = -1;
			if (posi != oldPosi) return true;
			else return false;
		}
		else if (IPC_PEN_DOWN && sprite >= 0){
			updateTouch();
			updateSprite(sprite, touchx - SQUARE_SIZE/2, touchy - SQUARE_SIZE/2);
		}
		else return false;
		swiWaitForVBlank();
	}
}

void rotateBoard(const bool isBoardRotated){
	BG3_XDY = 0;
	BG3_YDX = 0;
	if (isBoardRotated){
		BG3_XDX = (u16)(-1 << 8);
		BG3_YDY = (u16)(-1 << 8);
		BG3_CX = 257 << 8;
		BG3_CY = 183 << 8;
	}
	else{
		BG3_XDX = (u16)(1 << 8);
		BG3_YDY = (u16)(1 << 8);
		BG3_CX = 0;
		BG3_CY = 0;
	}
}

/**
 * Draws a line and move a sprite from (x1, y1) to (x2, y2)
 * 
 * @var animate if true, animate line and move sprite
 */ 
void drawLine(const int x1, const int y1, const int x2, const int y2, 
	const unsigned short color, bool animate=true){

    int yStep = SCREEN_WIDTH;
    int xStep = 1;
    int xDiff = x2 - x1;
    int yDiff = y2 - y1;
    const int s = getDragSprite(x1, y1);
    
    int errorTerm = 0;
    int offset = y1 * yStep + x1;
    int i;

    if (yDiff < 0){
		yDiff = -yDiff;
		yStep = -yStep;
    }

    if (xDiff < 0){
		xDiff = -xDiff;
		xStep = -xStep;
    }
    if (xDiff > yDiff){
		for (i = 0; i < xDiff + 1; i++){
			if (animate && i % 4 == 0){
				swiWaitForVBlank();
				updateSprite(s, (offset % 512) - 8, (offset / SCREEN_WIDTH) - 8);
			}
			BG_GFX[offset] = color|1<<15;
			BG_GFX[offset+1] = color|1<<15;
			BG_GFX[offset+yStep] = color|1<<15;
			offset += xStep;
			errorTerm += yDiff;
			if (errorTerm > xDiff){
				errorTerm -= xDiff;
				offset += yStep;
			}
		}
	}//end if xdiff > ydiff

	else{
		for (i = 0; i < yDiff + 1; i++){
			if (animate && i % 4 == 0){
				swiWaitForVBlank();
				updateSprite(s, (offset % 512) - 8, (offset / SCREEN_WIDTH) - 8);
			}
			BG_GFX[offset] = color|1<<15;
			BG_GFX[offset+1] = color|1<<15;
			BG_GFX[offset+yStep] = color|1<<15;
			offset += yStep;
			errorTerm += xDiff;
			if (errorTerm > yDiff){
				errorTerm -= yDiff;
				offset += xStep;  
			}
		}
	}
}

#define ABS(x) (((x)<0)?(-(x)):(x))
/**
 * Draws a thick ugly green line from the outgoing to the end field.
 * If the move is a capture, add the piece to the captured pieces
 */
void showMove(const int move, const bool isBoardRotated) {
	Point from = calcBoardxy(SQUARE_TO_64(MOVE_FROM(move)), isBoardRotated);
	from.x += SQUARE_SIZE/2; from.y += SQUARE_SIZE/2;
	Point to = calcBoardxy(SQUARE_TO_64(MOVE_TO(move)), isBoardRotated);
	to.x += SQUARE_SIZE/2; to.y += SQUARE_SIZE/2;
	const unsigned int color = RGB15(0, 20, 0);
	
	if (ABS(from.x - to.x) == 2*SQUARE_SIZE && ABS(from.y - to.y) == SQUARE_SIZE){
		drawLine (from.x, from.y, from.x + (to.x - from.x)/2, from.y, color);
		drawLine (from.x + (to.x - from.x)/2, from.y, to.x, to.y, color);
	}
	else if (ABS(from.y - to.y) == 2*SQUARE_SIZE && ABS(from.x-to.x) == SQUARE_SIZE){
		drawLine (from.x, from.y, from.x, from.y + (to.y - from.y)/2, color);
		drawLine (from.x, from.y + (to.y - from.y)/2, to.x, to.y, color);
	}	
	else drawLine (from.x, from.y, to.x, to.y, color);
	playSound(&put);
}

/**
 * Draws the square (x,y) x (x + SQUARE_SIZE, y + SQUARE_SIZE)
 * in the given color
 */
void drawSquare (const int x, const int y, const uint16 color){
	for (int j=y; j < y + SQUARE_SIZE; j++)
		for (int i=x; i < x + SQUARE_SIZE; i++)
			BG_GFX[j*SCREEN_WIDTH + i] = color;
}

/**
 * Draws the chess board squares
 */
void hideMove() {
	//slowest and badest possibility to hide the move
	swiWaitForVBlank();
	for (int j=0; j < 8; j++){
		for (int i=0; i < 8; i++){
			uint16 color;
			if ((i + j) % 2 == 1) color = 0x1A7F|1<<15;
			else color = 0x6739|1<<15;
			drawSquare (BOARD_LEFT+i*SQUARE_SIZE, BOARD_TOP+j*SQUARE_SIZE, color);
		}
	}
//	dmaCopy(board_gfx, BG_GFX, 24576*2);
}

/**
 * Highlights the given square
 */
void highlightField (const int square64, const bool suggest, const bool boardRotation){
	Point p = calcBoardxy(square64, boardRotation);
	p.x += 2; p.y += 1;
	
	for (int i=0; i < SQUARE_SIZE - 4; i++){
		p.y++;
		for (int j = p.x; j < p.x + SQUARE_SIZE - 4; j++){
			if (suggest && ((p.y + j) % 2 == 1)) BG_GFX [p.y*SCREEN_WIDTH+ j] = RGB15(10,0,60)|(1<<15);
			else if (!suggest && (p.y + j) % 2 == 0) 
				BG_GFX [p.y*SCREEN_WIDTH + j] =  RGB15(30,30,30)|(1<<15);
		}
	}
}

/**
 * Clears the screen and redraw the board
 */
void drawBoard (const bool isBoardRotated){
	clearScreen(BG_GFX, RGB15(31,31,31));
	
	hideMove();
	setbgColor(RGB15(31,31,31)|1<<15);
	if (!isBoardRotated){
		for (int i=0; i<8; i++){
			printChar(BG_GFX, BOARD_LEFT-8, BOARD_TOP + SQUARE_SIZE/2-4+ SQUARE_SIZE*i, 
				'8'-i, RGB15(10,10,10));
			printChar(BG_GFX, BOARD_LEFT+SQUARE_SIZE/2-3 + SQUARE_SIZE*i, 
				BOARD_TOP + SQUARE_SIZE*8+2, 'A'+i, RGB15(10,10,10));
		}
	}
	else{
		for (int i=0; i<8; i++){
			printChar(BG_GFX, BOARD_LEFT-8, BOARD_TOP + SQUARE_SIZE/2-4+ SQUARE_SIZE*i, 
				'1'+i, RGB15(10,10,10));
			printChar(BG_GFX, BOARD_LEFT+SQUARE_SIZE/2-3 + SQUARE_SIZE*i, 
				BOARD_TOP + SQUARE_SIZE*8+2, 'H'-i, RGB15(10,10,10));
		}
	}
}

#define insideBox(x1,y1,x2,y2,x,y) (x>=x1 && x<=x2 && y>=y1 && y<=y2)

/**
 * Draws a box with the given head and text. The user has to choose
 * between alt0 and alt1 with the stylus. 
 * The function blocks until the decision was made.
 * 
 * @return 1 for *alt1 and 0 for *alt0
 */

int drawBox2Alt (int x1, int y1, int x2, int y2, const char *head, 
	const char *str, const char *alt0, const char *alt1){
	
	const static int margin = 10;
	const static int xbuttonMargin = 20;
	const static int ybuttonMargin = 20;
	const static int noLength = getBigStringWidth(alt0);
	const static int yesLength = getBigStringWidth(alt1);
	
	const static uint16 bgColor = RGB15(20,20,20)|1<<15;
	const static uint16 marginColor = RGB15(25,25,25)|1<<15;
	const static uint16 marginColor2 = RGB15(15,15,15)|1<<15;
		
	//Draw a nice box
	clearRect(BG_GFX, x1, y1, x2, y2, bgColor);
	drawLine(x1, y1, x2, y1, marginColor, false);
	drawLine(x1, y2, x2, y2, marginColor2, false);
	drawLine(x1, y1, x1, y2, marginColor, false);
	drawLine(x2, y1, x2, y2, marginColor2, false);
	
	setbgColor(bgColor);
	int l = getStringWidth (head);
	int start = (x2-x1-l)/2;
	printString(BG_GFX, x1 + start, y1 + 5, head, RGB15(0,0,0));
	printString(BG_GFX, x1 + margin, y1 + 20, str, RGB15(0,0,0), false);
	printBigString(BG_GFX, x1 + xbuttonMargin, y2 - ybuttonMargin, alt1, RGB15(31,0,0));
	printBigString(BG_GFX, x2 - noLength - xbuttonMargin, y2 - ybuttonMargin, alt0, RGB15(31,0,0));

	u16 pen_prev = IPC_PEN_DOWN;
	u16 pen_curr = IPC_PEN_DOWN;
	int selected = -1;
	for (;;) {
		pen_prev = pen_curr;
		pen_curr = (IPC->buttons & IPC_PEN_DOWN);
    	if (PEN_HIT && selected == -1) {
	   		updateTouch();
      		if (insideBox (x1 + xbuttonMargin - 5, y2 - ybuttonMargin - 5, 
      		x1 + xbuttonMargin + yesLength + 5, y2 - ybuttonMargin + BIGFONTHEIGHT + 5, touchx, touchy)) {
      			selected = 1;
				printBigString(BG_GFX, x1 + xbuttonMargin, y2 - ybuttonMargin, alt1, RGB15(0,0,0));
      		}
      		if (insideBox (x2 - xbuttonMargin - noLength - 5, y2 - ybuttonMargin - 5, 
      		x2 - xbuttonMargin + 5, y2 - ybuttonMargin + BIGFONTHEIGHT + 5, touchx, touchy)) {
      			selected = 0;
				printBigString(BG_GFX, x2 - noLength - xbuttonMargin, y2 - ybuttonMargin, alt0, RGB15(0,0,0));
      		}
    	}
    	if (PEN_RELEASED && selected != -1) {
			drawBoard(true);
    		return selected;
    	}
		swiWaitForVBlank();
	}
}

/**
 * Draws a box with the given head and text. The user has to choose
 * Yes or No with the stylus. The function blocks until the decision was
 * made.
 * 
 * @return 1 for "yes" and 0 for "no"
 */
int drawBoxYesNo (int x1, int y1, int x2, int y2, const char *head, const char *str){
	return drawBox2Alt (x1, y1, x2, y2, head, str, "No", "Yes");
}

/**
 * Inits the gfx, assumes that a new game starts
 */
void init_board (const bool isBoardRotated) {
	drawBoard(isBoardRotated);
}

void init_gfx(){
	initSprites();

	BG3_CR = BG_BMP16_256x256;
	SUB_BG3_CR = BG_BMP16_256x256;
	
	SUB_BG3_XDY = 0;
	SUB_BG3_YDX = 0;
	SUB_BG3_XDX = (u16)(1 << 8);
	SUB_BG3_YDY = (u16)(1 << 8);
	SUB_BG3_CX = 0;
	SUB_BG3_CY = 0;
	
	rotateBoard(false);
	drawBoard(false);
}
