
#include <nds.h>
#include <cstdarg>
//#include <cstdio>
#include <cstdlib>
#include <cstring>
#include "fat.h"

#include "board.h"
#include "book.h"
#include "eval.h"
#include "fen.h"
#include "material.h"
#include "move.h"
#include "move_do.h"
#include "move_legal.h"
#include "option.h"
#include "pawn.h"
#include "posix.h"
#include "protocol.h"
#include "pst.h"
#include "search.h"
#include "trans.h"
#include "util.h"
#include "chessgfx.h"
#include "ndsMenu.h"
#include "fonts.h"
#include "check_raw.h"
#include "mate_raw.h"
#include "timesup_raw.h"
#include "san.h"
#include "clock.h"
#include "topScreen.h"

// constants

#define VERSION "2.1"
#define BOARDROTATED(colorI) (colorI.getSelection() == 1)
#define WHITETOMOVE(board) (board->sp % 2 == 0)

static const double NormalRatio = 1.0;
static const double PonderRatio = 1.25;


TransferSoundData check = {
		check_raw,		/* Sample address */
		check_raw_size,	/* Sample length */
		16000,				/* Sample rate */
		100,				/* Volume */
		64,					/* panning */
		2 					/* format */
	};

TransferSoundData mate = {
		mate_raw,		/* Sample address */
		mate_raw_size,	/* Sample length */
		16000,				/* Sample rate */
		100,				/* Volume */
		64,					/* panning */
		2 					/* format */
	};

TransferSoundData timesup = {
		timesup_raw,		/* Sample address */
		timesup_raw_size,	/* Sample length */
		16000,				/* Sample rate */
		100,				/* Volume */
		64,					/* panning */
		2 					/* format */
	};

// variables


static bool Searching; // search in progress?
static bool Infinite; // infinite or ponder mode?
static bool Delay; // postpone "bestmove" in infinite/ponder mode?

// prototypes

static void init              ();
static void problem_loop_step();
static void loop_step2         ();

static void parse_go          (char string[]);

static void send_best_move    ();

static bool string_equal      (const char s1[], const char s2[]);

extern void refreshScreen (bool repaint);
extern void stopSearch();

#define COMPUTER 1
#define ME 0
#define OTHERHUMAN 1

board_t inputBoard[1];
uint8 player = ME;
bool ponderhit = false;
int pondermove = MoveNull;
ChessClock clockW;
ChessClock clockB;

int captured_pieces[32];
int capture_count=0;


SimpleSelectionMenuItem colorI("Color", "Color of the player");
SimpleSelectionMenuItem fixedTimeI("Movetime", "The maximal time for the computer to make a move");
SimpleSelectionMenuItem fixedDepthI("Depth", "The maximal Depth the computer calculates");
SimpleSelectionMenuItem varTimeI("Time", "The time for both players");
SimpleSelectionMenuItem levelI("Level", "Easy playing modes");
SimpleSelectionMenuItem openingBookI("Opening book", "");
SimpleSelectionMenuItem ponderI("Pondermode", "If switched on, the computer calculates in also in the players time");
SimpleSelectionMenuItem debugI("Debug", "If switched on, the DS will output some debug msg.");

class ForceItem:public SimpleMenuItem{
	public:
	ForceItem(char * _name, char * _help):SimpleMenuItem(_name, _help){}
	int execute(uint16 keys){
		stopSearch();
		return 1;
	}
};

class SuggestItem:public SimpleMenuItem{
	public:
	SuggestItem(char * _name, char * _help):SimpleMenuItem(_name, _help){}
	int execute(uint16 keys){
		int move = SearchBest->pv[1];
		if (pondermove != MoveNone) move = pondermove;
		if (move == MoveNone){
			//look in the opening book
			move = book_move(inputBoard);
		}
		if (move == MoveNone) return 0;
		highlightField(SQUARE_TO_64(MOVE_FROM(move)), true, BOARDROTATED(colorI));
		highlightField(SQUARE_TO_64(MOVE_TO(move)), true, BOARDROTATED(colorI));
		return 1;
	}
};

class UndoItem:public SimpleMenuItem{
	public:
	const static int MAXMOVES = 256;
	undo_t undo[MAXMOVES];
	int moves[MAXMOVES];
	int lastmove;
	bool lastWasUndo;
	UndoItem(char * _name, char * _help):SimpleMenuItem(_name, _help), lastmove(0), lastWasUndo(false), undoMove(false){}
	
	void moveDo(int move){
		moves[lastmove] = move;
		if (move_is_capture(move, inputBoard)){
			captured_pieces[capture_count++] = move_capture(move, inputBoard);
		}
		move_do(inputBoard, move, &undo[lastmove]);
		//move_do(SearchInput->board, move, &undo[lastmove]);
		lastmove++;
		lastWasUndo = false;
	}
	
	void reset(){
		lastmove = 0;
	}
	
	/**
	 * Checks if the player wants to undo the last move and if so, undo it
	 * @returns true if the last move was undone, else false
	 */
	bool check4undo(){
		if (undoMove){
			move_undo(inputBoard, moves[lastmove-1], &undo[lastmove-1]);
			move_undo(SearchInput->board, moves[lastmove-1], &undo[lastmove-1]);
			if (move_is_capture(moves[lastmove-1], inputBoard)) capture_count--;
			lastmove--;
			undoMove = false;
			lastWasUndo = true;
			ponderhit = false;
			hideMove();
			return true;
		}
		return false;
	}
	
	int execute(uint16 keys){
		if (undoMove == false){
			undoMove = true;
			//stop searching or pondering
			stopSearch();
	      	ponderhit = false;
	      	//Change Player, no ponderhit
	      	if (player == 0) player = 1;
	   	  	else player = 0;
		}
		return 1;
	}
	
	private:
		bool undoMove;
	
};

UndoItem undoI("Undo move", "Undoes the last move");
Menu inGameMenu("Game");
ForceItem forceI("Force move", "Forces the computer to make a move");
SuggestItem suggestI("Suggest move", "The computer suggests a move for you");
selectionMenu timeSettings("CPU settings");
subMenuItem timeSettingsI("CPU settings", "", &timeSettings);
Menu newGameMenu("New game");
Menu optionMenu("Game settings");
subMenuItem newGameMenuI("New game", "", &newGameMenu);
subMenuItem optionMenuI("Settings", "", &optionMenu);

class NewGameItem:public SimpleMenuItem{
	public:
	bool newGame;
	NewGameItem(char * _name, char * _help):SimpleMenuItem(_name, _help), newGame(false){}
	int execute(uint16 keys){
	  	//Stop searching or pondering
	  	stopSearch();
      	newGame = true;
		return 2;
	}
	
	bool check4newGame(){
		if (newGame){
			newGame = false;
			option_init();
			//Change options if level-play is selected
			if (timeSettings.getChoosen() == &levelI){
				adjustLevelSettings(levelI.getSelection() + 1);
			}
			
			clockW.reset();
			clockB.reset();
			board_from_fen(SearchInput->board,StartFen);
			board_from_fen(inputBoard,StartFen);
			capture_count = 0;
			undoI.reset();
			trans_clear(Trans);
			search_clear();
			//Reset suggest move
			SearchBest->pv[1] = MoveNone;
			resetLastMove();

			pawn_delete();
			pawn_init();
			pawn_alloc();

			material_delete();
			material_init();
			material_alloc();

			pst_init();
			eval_init();

			if (colorI.getSelection() != 1) player = ME;
			else player = COMPUTER;
			swiWaitForVBlank();
			init_board(BOARDROTATED(colorI));
			fill_board(inputBoard, BOARDROTATED(colorI));
			drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
			refreshScreen(true);
			hideMove();
			return true;
		}
		return false;
	}
	
	private:
		
		void adjustLevelSettings(const int l) {
			//Set time to three seconds
			//timeSettings.setChoosen(&fixedTimeI);
			//fixedTimeI.setSelection(1);
			//Switch on pondering
			ponderI.setSelection(0);
			//Turn on opening book
			openingBookI.setSelection(0);
			option_set("OwnBook", "true");
			
			
			if (l < 11) {
				//Switch off pondering
				ponderI.setSelection(1);
			}
			if (l < 10) {
				//Set time to one second
				//timeSettings.setChoosen(&fixedTimeI);
				//fixedTimeI.setSelection(0);				
			}
			if (l < 9) {
				//Turn off History and NullMove pruning
				option_set("History Pruning", "false");
				option_set("NullMove Pruning", "Never");
				option_set("Verification Search", "Never");
			}
			if (l < 8) {
				//Set search depth to one 
				//timeSettings.setChoosen(&fixedDepthI);
				//fixedDepthI.setSelection(1);
			}
			if (l < 7) {
				//Opening book on
				//timeSettings.setChoosen(&fixedDepthI);
				//fixedDepthI.setSelection(0);
			}
			if (l < 6) {
				//Quiescence Check Plies on
				//Turn off opening book
				openingBookI.setSelection(1);
				option_set("OwnBook", "false");
				option_set("King Safety", "100");
			}

			if (l < 5) {
				option_set("Quiescence Check Plies", "0");
				option_set("Material", "100");
				option_set("Pawn Structure", "100");
				option_set("King Safety", "50");
			}
			
			if (l < 4) {
				option_set("Material", "70");
				option_set("Pawn Structure", "70");
				option_set("King Safety", "30");
			}
			
			if (l == 2) {
				option_set("Material", "30");
				option_set("Pawn Structure", "30");
			}
			
			if (l == 1) {
				option_set("Material", "10");
				option_set("Pawn Structure", "10");
				option_set("King Safety", "20");
			}
		}
};
NewGameItem startGameI("Start game!", "Starts a new game");

class BackItem:public SimpleMenuItem{
	public:
	BackItem(char * _name, char * _help):SimpleMenuItem(_name, _help){}
	int execute(uint16 keys){
		return 1;
	}
	
};
BackItem backI("Back...", "Back to previous Menu");

static const int FixedTimes[] = {1,3,5,10,15,20,40,-1};

void initMenu(){
	static bool init = false;
	if (!init){
		//inGameMenu.setPosition (50, 30);
		inGameMenu.addMenuItem (&undoI);
		inGameMenu.addMenuItem (&suggestI);
		inGameMenu.addMenuItem (&forceI);
		optionMenu.addMenuItem(&timeSettingsI);		
		inGameMenu.addMenuItem(&optionMenuI);
		inGameMenu.addMenuItem(&newGameMenuI);
		
		debugI.addSelectionItem("on");
		debugI.addSelectionItem("off");
		debugI.setSelection(1);
		timeSettings.addMenuItem(&fixedTimeI);
		fixedTimeI.addSelectionItem("1s");
		fixedTimeI.addSelectionItem("3s");
		fixedTimeI.addSelectionItem("5s");
		fixedTimeI.addSelectionItem("10s");
		fixedTimeI.addSelectionItem("15s");
		fixedTimeI.addSelectionItem("20s");
		fixedTimeI.addSelectionItem("40s");
		fixedTimeI.addSelectionItem("Inf.");
		fixedTimeI.setSelection(1);
		
		timeSettings.addMenuItem(&fixedDepthI);
		fixedDepthI.addSelectionItem("1");
		fixedDepthI.addSelectionItem("2");
		fixedDepthI.addSelectionItem("3");
		fixedDepthI.addSelectionItem("4");
		fixedDepthI.addSelectionItem("5");
		fixedDepthI.addSelectionItem("6");
		fixedDepthI.addSelectionItem("7");
		fixedDepthI.addSelectionItem("8");
		fixedDepthI.addSelectionItem("9");
		fixedDepthI.setSelection(5);
		
		timeSettings.addMenuItem(&varTimeI);
		varTimeI.addSelectionItem("5m");
		varTimeI.addSelectionItem("10m");
		varTimeI.addSelectionItem("20m");
		varTimeI.addSelectionItem("40m");
		varTimeI.addSelectionItem("5m+1s");
		varTimeI.addSelectionItem("10m+2s");
		varTimeI.addSelectionItem("20m+2s");
		varTimeI.addSelectionItem("40m+3s");
		varTimeI.setSelection(6);
		timeSettings.addMenuItem(&levelI);
		levelI.addSelectionItem("1");
		levelI.addSelectionItem("2");
		levelI.addSelectionItem("3");
		levelI.addSelectionItem("4");
		levelI.addSelectionItem("5");
		levelI.addSelectionItem("6");
		levelI.addSelectionItem("7");
		levelI.addSelectionItem("8");
		levelI.addSelectionItem("9");
		levelI.addSelectionItem("10");
		levelI.addSelectionItem("11");
		levelI.setSelection(0);
		
		timeSettings.setChoosen(&fixedTimeI);
		timeSettings.addMenuItem(&backI);
		
		//timeSettings.setPosition(10, 10);
		//newGameMenu.setPosition(10, 10);
		optionMenu.addMenuItem(&openingBookI);
		openingBookI.addSelectionItem("on");
		openingBookI.addSelectionItem("off");
		openingBookI.setSelection(0);
		
		optionMenu.addMenuItem(&ponderI);
		ponderI.addSelectionItem("on");
		ponderI.addSelectionItem("off");
		ponderI.setSelection(0);
		
		newGameMenu.addMenuItem(&colorI);
		colorI.addSelectionItem("white");
		colorI.addSelectionItem("black");
		colorI.addSelectionItem("H vs. H");
		colorI.setSelection(0);
		
		newGameMenu.addMenuItem(&startGameI);
		optionMenu.addMenuItem(&debugI);
		optionMenu.addMenuItem(&backI);
		
		init = true;
	}
}

void setClocks() {
	if (timeSettings.getChoosen() == &varTimeI){
		int inc = 0;
		uint32 time = 5;
		if (varTimeI.getSelection() == 1 || varTimeI.getSelection() == 5) time = 10;
		else if (varTimeI.getSelection() == 2 || varTimeI.getSelection() == 6) time = 20;
		else if (varTimeI.getSelection() == 3 || varTimeI.getSelection() == 7) time = 40;

		if (varTimeI.getSelection() == 4) inc = 1;
		else if (varTimeI.getSelection() == 5) inc = 2;
		else if (varTimeI.getSelection() == 6) inc = 2;
		else if (varTimeI.getSelection() == 7) inc = 3;
		clockW.set(true, time*60, inc);
		clockB.set(true, time*60, inc);
	}
	else {
		clockW.set(false);
		clockB.set(false);		
	}
}

bool isDebug(){
	return !debugI.getSelection();
}

/**
 * Saves the current game to a file
 */
void saveBoard(const char* fname) {
	if (hasFAT()){
		FILE* saveFile = fopen(fname, "w+");
		if (saveFile != NULL) {
			fprintf(saveFile, "%d %d ", player, colorI.getSelection());
			
			fprintf(saveFile, "%u %u ", clockW.getElapsedSeconds(), clockB.getElapsedSeconds());
			/*fprintf(saveFile, "%u %u %u %u ", clockW.start_real, clockW.start_cpu, clockW.elapsed_real, clockW.elapsed_cpu); 
			fprintf(saveFile, "%u %u %u %u ", clockB.start_real, clockB.start_cpu, clockB.elapsed_real, clockB.elapsed_cpu);*/
			
			//Save time settings
			fprintf(saveFile, "%u ", timeSettings.getChoosenID());
			SimpleSelectionMenuItem *item = static_cast<SimpleSelectionMenuItem *>(timeSettings.getChoosen());
			int t = item->getSelection();
			fprintf(saveFile, "%u ", t);
			
			//Save moves
			for (int i=0; i<undoI.lastmove; i++) {
				fprintf(saveFile, " %d", undoI.moves[i]);
			} 
			fclose(saveFile);
		}
	}
}

bool loadBoard(const char* fname) {
	if (hasFAT()){
		FILE* saveFile = fopen(fname, "rb");
		if (saveFile != NULL) {
			int color;
			int pl = ME;
			fscanf(saveFile, "%d %d", &pl, &color);
			player = pl;
			colorI.setSelection(color);
			//TODO
			int cw, cb;
			fscanf(saveFile, "%u %u", &cw, &cb);
			clockW.setElapsedSeconds(cw);
			clockB.setElapsedSeconds(cb);
			/*fscanf(saveFile, "%u%u%u%u", &clockW.start_real, &clockW.start_cpu, &clockW.elapsed_real, &clockW.elapsed_cpu); 
			fscanf(saveFile, "%u%u%u%u", &clockB.start_real, &clockB.start_cpu, &clockB.elapsed_real, &clockB.elapsed_cpu);*/
			//Load time settings
			int t;
			fscanf(saveFile, "%u", &t);
			timeSettings.setChoosenID(t);
			fscanf(saveFile, "%u", &t);
			SimpleSelectionMenuItem *item = static_cast<SimpleSelectionMenuItem *>(timeSettings.getChoosen());
			item->setSelection(t);
			//Load moves
			while (!feof(saveFile)) {
				int move;
				undo_t undo;
				if (fscanf(saveFile, "%d", &move) == 1) {
					undoI.moveDo(move);
					move_do(SearchInput->board, move, &undo);
				}
			} 
			fclose(saveFile);
			return true;
		}
	}
	return false;
}

/**
 * Checks whether a file with the given fname exists
 */
bool saveExists(const char* fname) {
	if (hasFAT()) {
		FILE* saveFile = fopen(fname, "rb");
		if (saveFile != NULL){
			fclose(saveFile); 
			return true;
		}
	}
	return false;
}

/**
 * Tells the engine to stop a search as soon as possible
 */
void stopSearch(){
	if (Searching) {
		ASSERT(Infinite);
		SearchInfo->stop = true;
		Infinite = false;
	} 
	else if (Delay) {
		send_best_move();
		Delay = false;
	}
}

void start_clock(const int player){
	if (WHITETOMOVE(inputBoard)) clockW.start();
	else clockB.start();
}

void stop_clock(const int player){
	clockW.stop();
	clockB.stop();
}


/**
 * Checks if the game has finished (time, mate, stalemate)
 */ 
bool isFinished(){
	if (clockW.isFinished() && !(undoI.lastmove==0)) return true;
	else if (clockB.isFinished() && !(undoI.lastmove==0)) return true;
	if (board_is_mate(inputBoard)) return true;
	else if (board_is_stalemate(inputBoard)) return true;
	else if (isSecondRepetition(inputBoard)) return true;
	else return false;
}

/**
 * Creates from the choosen options a goString
 */
void getGoString(char * output, const bool ponder){
	if (timeSettings.getChoosen() == &fixedTimeI){
		const int time = FixedTimes[fixedTimeI.getSelection()] * 1000;
		if (time < 0) sprintf(output, "go infinite");
		else sprintf(output, "go movetime %d", time);
	}
	else if (timeSettings.getChoosen() == &fixedDepthI){
		const int depth = 1 + fixedDepthI.getSelection();
		sprintf(output, "go depth %d", depth);
	}
	else if (timeSettings.getChoosen() == &levelI) {
		if (levelI.getSelection() + 1 < 7) sprintf(output, "go depth 1");
		else if (levelI.getSelection() + 1 < 9) sprintf(output, "go depth 2");
		else if (levelI.getSelection() + 1 < 10){
			const int time = FixedTimes[0] * 1000;
			sprintf(output, "go movetime %d", time);
		}
		else {
			const int time = FixedTimes[1] * 1000;
			if (ponder) sprintf(output, "go movetime %d ponder", time);
			else sprintf(output, "go movetime %d", time);
		}
		return;
	}
	else if (timeSettings.getChoosen() == &varTimeI){
		sprintf(output, "go btime %d binc %d wtime %d winc %d", 
				clockB.getSeconds()*1000, clockB.inc*1000, clockW.getSeconds()*1000, clockW.inc*1000);
	}
	if (ponder) strcat(output, " ponder");
}

void check4Sounds() {
	if (board_is_check(inputBoard)) playSound(&check);
	if (board_is_mate(inputBoard)) playSound(&mate);
}

static bool newMove = false;
static int screen = 0;
void refreshScreen(bool repaint) {
	if (screen == 0) {
		refreshScreen (repaint, newMove, player, isDebug(), inputBoard, &clockW, &clockB, SearchBest, 
			undoI.moves, undoI.undo);
	}
	else if (screen == 1) {
		if (undoI.lastWasUndo) repaint = true;
		refreshScreen2 (repaint, newMove, player, isDebug(), inputBoard, &clockW, &clockB, SearchBest, 
			undoI.moves, undoI.undo);
		undoI.lastWasUndo = false;
	}
	newMove = false;
}
/**
 * Keyhandler
 */ 
bool checkKeys(){
	bool repaint = false;
	static uint16 lastKeys = 0;
	uint16 currKeys = ~(REG_KEYINPUT);
	
	if (!(lastKeys & KEY_START) && (currKeys & KEY_START)){
		REG_POWERCNT &= ~POWER_2D_A;
		if (player == COMPUTER){
			forceI.active = true;
			suggestI.active = false;
		}
		else{
			forceI.active = false;
			suggestI.active = true;
		}
		if (undoI.lastmove == 0) undoI.active = false;
		else undoI.active = true;
		bool w = clockW.isRunning();
		bool b = clockB.isRunning();
		bool t = SearchCurrent->timer->running;
		if (t) my_timer_stop(SearchCurrent->timer);
		clockW.stop(false); clockB.stop(false);
		inGameMenu.display();
		setClocks();
		repaint = true;
		if (t) my_timer_start(SearchCurrent->timer);
		if (w) clockW.start();
		if (b) clockB.start();
		REG_POWERCNT |= POWER_2D_A;
		if (openingBookI.getSelection() == 1) option_set("OwnBook", "false");
		else option_set("OwnBook", "true");
	}
	else if (!(lastKeys & KEY_L) && (currKeys & KEY_L)){
		if (undoI.lastmove > 0) undoI.execute(KEY_A);
	}
	else if (!(lastKeys & KEY_R) && (currKeys & KEY_R)){
		if (player == ME) suggestI.execute(KEY_A);
	}
	else if (!(lastKeys & KEY_A) && (currKeys & KEY_A)){
		if (player == COMPUTER) forceI.execute(KEY_A);
	}
	else if (!(lastKeys & KEY_SELECT) && (currKeys & KEY_SELECT)){
		screen = (screen+1) %2;
		repaint = true;
	}
	
	lastKeys = currKeys;
	return repaint;
}	

static bool breakComputerSearch = false;

/**
 * Looks for input from a human player,
 * from is either ME or OTHERHUMAN
 */
void lookforInput(const int from){
	bool repaint = false;
	static uint8 pause = 0;
	//If searching this function will be called ~2000 times/sec
	if (Searching && pause++ < 100) return;
	pause = 0;

	static int lastMode = 0;
	if (lastMode != colorI.getSelection()) {
		//User has changed setting
		init_board(BOARDROTATED(colorI));
		fill_board(inputBoard, BOARDROTATED(colorI));
		drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
		hideMove();
		if (colorI.getSelection() == 2) {
			stopSearch();
			ponderhit = false;
			breakComputerSearch = true;
		}
		else if (player == ME && WHITETOMOVE(inputBoard) && colorI.getSelection() == 1){
			player = COMPUTER;
			stopSearch(); // Pondering
			ponderhit = false;
		}
		else if (player == ME && !WHITETOMOVE(inputBoard) && colorI.getSelection() == 0){
			player = COMPUTER;
			stopSearch(); // Pondering
			ponderhit = false;
		}
		else if (player == COMPUTER && WHITETOMOVE(inputBoard) && colorI.getSelection() == 0){
			player = ME;
			stopSearch();
			breakComputerSearch = true;
		}
		else if (player == COMPUTER && !WHITETOMOVE(inputBoard) && colorI.getSelection() == 1){
			player = ME;
			stopSearch();
			breakComputerSearch = true;
		}
		//}
		lastMode = colorI.getSelection();
		return;
	}
	int other;
	
	repaint = checkKeys();
	refreshScreen (repaint);
	if (colorI.getSelection() != 2 && player == COMPUTER) return;
	if (isFinished()) return;
	
	if (waitT(inputBoard, BOARDROTATED(colorI))){
		stop_clock(from);
		hideMove();
		if ((!isFinished()) && pondermove == getLastMove()){
			//ponderhit, the engine continues searching until the time is over
			ponderhit = true;
			if (Searching) {
				ASSERT(Infinite);
				SearchInput->infinite = false;
				Infinite = false;
			} 
			else if (Delay) {
				send_best_move();
				Delay = false;
			}
			//printf("\x1b[12;0HPonder!!");
		}
		else{
			ponderhit = false;
			stopSearch();
		}
		if (player == ME) player = COMPUTER;
		else player = ME;
		undoI.moveDo(getLastMove());
		fill_board(inputBoard, BOARDROTATED(colorI));
		drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
		saveBoard("chess.save");
		newMove = true;
		check4Sounds();		
		start_clock(other);
	}
}


void loop() {
	Searching = false;
	Infinite = false;
	Delay = false;

	search_clear();
	init();

	board_from_fen(SearchInput->board,StartFen);
	board_from_fen(inputBoard,StartFen);
//	board_from_fen(SearchInput->board, "8/8/B1nK4/B7/k7/2Q5/8/1b6 w - -");
//	board_from_fen(inputBoard, "8/8/B1nK4/B7/k7/2Q5/8/1b6 w - -");
//	board_from_fen(inputBoard, "5Q2/p1K5/8/rpp5/2p2p2/2R2r2/2R5/1k6 w - -");
//	board_from_fen(SearchInput->board, "5Q2/p1K5/8/rpp5/2p2p2/2R2r2/2R5/1k6 w - -");
	//loadBoard(SearchInput->board, "chess.save");
	if (saveExists("chess.save")) {
		if (drawBoxYesNo(40, 50, 220, 120, "Save game", "Do you like to continue\nyour last game?") == 1)
			loadBoard("chess.save");
	}
	init_board(BOARDROTATED(colorI));
	fill_board(SearchInput->board, BOARDROTATED(colorI));
	drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
	setClocks();
	refreshScreen(true);
	while (true) loop_step2();
//	while (true) problem_loop_step();
}

// init()

static void init() {
	static bool init = false;
	if (!init) {
		init = true;

		if (option_get_bool("OwnBook"))
			book_open(option_get_string("BookFile"));

		trans_alloc(Trans);

		pawn_init();
		pawn_alloc();

		material_init();
		material_alloc();

		pst_init();
		eval_init();
		initMenu();
		clockB.reset();
		clockW.reset();
	}
}

// event()
void event() {
}

static void problem_loop_step(){
  	undo_t undo[1];
	static char string[1024]; //was 1024
	ponderhit = false;
	pondermove = MoveNone;

	if (isFinished()){
		clockW.stop(false); clockB.stop(false);
	}
	while (isFinished()){
		lookforInput(ME);
		swiWaitForVBlank();
		if (startGameI.check4newGame()) return;
		if (undoI.check4undo()){
			start_clock(player);
			fill_board(inputBoard, BOARDROTATED(colorI));
			newMove = true;
			return;
		}
	}
	
	stop_clock(COMPUTER);
	//start clock after the first move
	if (undoI.lastmove > 0) start_clock(ME);
	while (player == ME){
		lookforInput(ME);
		if (startGameI.check4newGame()) return;
		swiWaitForVBlank();
	}
	hideMove();
	if (!undoI.check4undo() && getLastMove() != MoveNone) undoI.moveDo(getLastMove());
	else{
		stop_clock(ME);
		start_clock(COMPUTER);
	}
	newMove = true;
	fill_board(SearchInput->board, BOARDROTATED(colorI));
	sprintf(string, "go mate %d",4);
	if (isFinished()) return;
	parse_go(string);
	//do blacks move
	if (!undoI.check4undo()){
		int move = SearchBest->move;
		undoI.moveDo(move);
		move_do(inputBoard, move, undo);
		showMove(move, BOARDROTATED(colorI));
	}
	player = ME;
	if (startGameI.check4newGame()) return;
	fill_board(SearchInput->board, BOARDROTATED(colorI));	
	newMove = true;
}

static char *string = (char *)my_malloc(1024*sizeof(char));

/**
 * Main Loop
 */
static void loop_step2(){
  	undo_t undo[1];
	ponderhit = false;
	pondermove = MoveNone;
	resetLastMove();
	
	if (isFinished()){
		clockW.stop(false); clockB.stop(false);
		
		while (isFinished()){
			lookforInput(ME);
			swiWaitForVBlank();
			if (startGameI.check4newGame()) return;
			if (undoI.check4undo()){
				start_clock(player);
				fill_board(inputBoard, BOARDROTATED(colorI));
				drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
				newMove = true;
				return;
			}
		}
	}
	
	//We need the condition SearchBest->value > -20000, otherwise there occurs a nifty error
	if (SearchBest->value > -20000 && player == ME && ponderI.getSelection() == 0 
		&& SearchBest->pv[1] != MoveNone && SearchBest->pv[0] != MoveNone
		&& colorI.getSelection() != 2 && 
		(timeSettings.getChoosen() != &levelI || (timeSettings.getChoosen() == &levelI && levelI.getSelection() == 10))){
		pondermove = SearchBest->pv[1];
		//Make the move on the ponderboard (SearchInput->board)
		move_do (SearchInput->board, pondermove, undo);
		getGoString(string, true);
		//printf(string);
		stop_clock(COMPUTER);
		start_clock(ME);
		//inside this white makes his move
		parse_go(string);
		if (startGameI.check4newGame()) return;
		//Special case: search ended without human player made his move
		while (player == ME) {
			lookforInput(ME);
			if (startGameI.check4newGame()) return;
			swiWaitForVBlank();			
		}
	}
	else if (player == ME){
		stop_clock(COMPUTER);
		//start clock after the first move
		if (undoI.lastmove > 0) start_clock(ME);
		while (player == ME){
			lookforInput(ME);
			if (startGameI.check4newGame()) return;
			swiWaitForVBlank();
		}
	}
	hideMove();
	
	//Human has made his move (it is already done on inputBoard and undo History)
	bool otherHumanMove = false;
	if (ponderhit){
		//SearchInput->board is already one move in front
		//Computer has made a ponder move, so we can skip calculation 
	}
	else{
		if (pondermove != MoveNone) move_undo (SearchInput->board, pondermove, undo);		
		//Make my move
		if (!undoI.check4undo() && getLastMove() != MoveNone) {
			move_do(SearchInput->board, getLastMove(), undo);
		}
		else{
			stop_clock(ME);
			start_clock(COMPUTER);
		}
		newMove = true;
		fill_board(SearchInput->board, BOARDROTATED(colorI));
		drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
		if (isFinished()) return;
		if (colorI.getSelection() == 2) {
			//Play against another human (black's move)
			while (colorI.getSelection() != 0 && player == OTHERHUMAN){
				lookforInput(OTHERHUMAN);
				if (startGameI.check4newGame()) return;
				swiWaitForVBlank();
			}
			if (colorI.getSelection() != 0) otherHumanMove = true;
		}
		
		if (!otherHumanMove){
			stop_clock(COMPUTER);
			breakComputerSearch = false;
			//Play against computer
			getGoString(string, false);
			int c=10;
			while(c-- > 0){
				lookforInput(0);
				swiWaitForVBlank();
			}
			start_clock(COMPUTER);
			//printf(string);
			parse_go(string);
		}
	}
	//do COMPUTERs or OTHERHUMANs move
	if (!undoI.check4undo()){
		if (otherHumanMove) {
			move_do(SearchInput->board, getLastMove(), undo);
		}
		else if (!breakComputerSearch) {
			undoI.moveDo(SearchBest->move);
			move_do(SearchInput->board, SearchBest->move, undo);
			showMove(SearchBest->move, BOARDROTATED(colorI));
		}
		check4Sounds();
	}
	player = ME;
	if (otherHumanMove && colorI.getSelection() == 1) player = COMPUTER;
	stop_clock(COMPUTER);
	saveBoard("chess.save");
	if (startGameI.check4newGame()) return;
	fill_board(SearchInput->board, BOARDROTATED(colorI));
	drawCapturedPieces(captured_pieces, capture_count, BOARDROTATED(colorI));
	newMove = true;
}

static void parse_go(char string[]) {

   const char * ptr;
   bool infinite, ponder;
   int depth, mate, movestogo;
   sint64 nodes;
   double binc, btime, movetime, winc, wtime;
   double time, inc;
   double time_max, alloc;

   // init

   infinite = false;
   ponder = false;

   depth = -1;
   mate = -1;
   movestogo = -1;

   nodes = -1;

   binc = -1.0;
   btime = -1.0;
   movetime = -1.0;
   winc = -1.0;
   wtime = -1.0;

   // parse

   ptr = strtok(string," "); // skip "go"

   for (ptr = strtok(NULL," "); ptr != NULL; ptr = strtok(NULL," ")) {

      if (false) {

      } else if (string_equal(ptr,"binc")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         binc = double(atoi(ptr)) / 1000.0;
         ASSERT(binc>=0.0);

      } else if (string_equal(ptr,"btime")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         sint32 time = atoi(ptr) / 1000;
         btime = time * 1.0;
         //btime = double(atoi(ptr)) / 1000.0;
         ASSERT(btime>=0.0);

      } else if (string_equal(ptr,"depth")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         depth = atoi(ptr);
         ASSERT(depth>=0);

      } else if (string_equal(ptr,"infinite")) {

         infinite = true;

      } else if (string_equal(ptr,"mate")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         mate = atoi(ptr);
         ASSERT(mate>=0);

      } else if (string_equal(ptr,"movestogo")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         movestogo = atoi(ptr);
         ASSERT(movestogo>=0);

      } else if (string_equal(ptr,"movetime")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         movetime = double(atoi(ptr)) / 1000.0;
         ASSERT(movetime>=0.0);

      } else if (string_equal(ptr,"nodes")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         nodes = my_atoll(ptr);
         ASSERT(nodes>=0);

      } else if (string_equal(ptr,"ponder")) {

         ponder = true;

      } else if (string_equal(ptr,"searchmoves")) {

         // dummy

      } else if (string_equal(ptr,"winc")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         winc = double(atoi(ptr)) / 1000.0;
         ASSERT(winc>=0.0);

      } else if (string_equal(ptr,"wtime")) {

         ptr = strtok(NULL," ");
         if (ptr == NULL) my_fatal("parse_go(): missing argument\n");

         wtime = double(atoi(ptr)) / 1000.0;
         ASSERT(wtime>=0.0);
      }
   }

   // init

   search_clear();

   // depth limit

   if (depth >= 0) {
      SearchInput->depth_is_limited = true;
      SearchInput->depth_limit = depth;
   } else if (mate >= 0) {
      SearchInput->depth_is_limited = true;
      SearchInput->depth_limit = mate * 2 - 1; // HACK: move -> ply
   }

   // time limit

   if (COLOUR_IS_WHITE(SearchInput->board->turn)) {
      time = wtime;
      inc = winc;
   } else {
      time = btime;
      inc = binc;
   }

   if (movestogo <= 0 || movestogo > 30) movestogo = 30; // HACK
   if (inc < 0.0) inc = 0.0;

   if (movetime >= 0.0) {

      // fixed time

      SearchInput->time_is_limited = true;
      SearchInput->time_limit_1 = movetime * 5.0; // HACK to avoid early exit
      SearchInput->time_limit_2 = movetime;

   } else if (time >= 0.0) {

      // dynamic allocation

      time_max = time * 0.95 - 1.0;
      if (time_max < 0.0) time_max = 0.0;

      SearchInput->time_is_limited = true;
//HACK!! for Dualis
//      alloc = (time_max) / double(movestogo);
      alloc = (time_max + inc*double(movestogo-1)) / double(movestogo);
      //alloc *= (option_get_bool("Ponder") ? PonderRatio : NormalRatio);
      alloc *= (ponderI.getSelection() == 0 ? PonderRatio : NormalRatio);
      if (alloc > time_max) alloc = time_max;
      SearchInput->time_limit_1 = alloc;

      alloc = (time_max + inc * double(movestogo-1)) * 0.5;
//      alloc = (time_max) * 0.5;
      if (alloc < SearchInput->time_limit_1) alloc = SearchInput->time_limit_1;
      if (alloc > time_max) alloc = time_max;
      SearchInput->time_limit_2 = alloc;
   }

   if (infinite || ponder) SearchInput->infinite = true;

   // search

   ASSERT(!Searching);
   ASSERT(!Delay);

   Searching = true;
   Infinite = infinite || ponder;
   Delay = false;

   search();
   search_update_current();

   ASSERT(Searching);
   ASSERT(!Delay);

   Searching = false;
   Delay = Infinite;

   if (!Delay) send_best_move();
}

// send_best_move()

static void send_best_move() {

   //char ponder_string[256];
   char move_string[256];
   int move;
   mv_t * pv;

   // info

   // HACK: should be in search.cpp

   const double time = SearchCurrent->time;
   const double speed = SearchCurrent->speed;
   const double cpu = SearchCurrent->cpu;
   const sint64 node_nb = SearchCurrent->node_nb;
   sprintf(move_string, "Nps:%.0f", node_nb/time);
   if (isDebug()) printString(BG_GFX_SUB, 5, 80, move_string, TEXT_COLOR);
   send("info time %.0f nodes " S64_FORMAT " nps %.0f cpuload %.0f",time,node_nb,speed,cpu*1000.0);

   trans_stats(Trans);
   // pawn_stats();
   // material_stats();

   // best move

   move = SearchBest->move;
   pv = SearchBest->pv;

   move_to_string(move,move_string,256);

   if (pv[0] == move && move_is_ok(pv[1])) {
      //move_to_string(pv[1],ponder_string,256);
      //send("bestmove %s ponder %s",move_string,ponder_string);
      send("bestmove %s",move_string);
   } else {
      send("bestmove %s",move_string);
   }
//   bestmove = move;
//   strcat (moves_string, " ");
//   strcat (moves_string, move_string);
}


// send()

void send(const char format[], ...) {
	if (isDebug()){
	   va_list arg_list;
		static char string[1024]; //was 4096

		ASSERT(format!=NULL);

		va_start(arg_list,format);
		vsprintf(string,format,arg_list);
		va_end(arg_list);

/*		printf("\x1b[15;0H                                                                                   ");
		printf("\x1b[15;0H%s\n",string);*/
		clearRect(BG_GFX_SUB, 0, 160, SCREEN_WIDTH, SCREEN_HEIGHT, BGCOLOR);
		printString (BG_GFX_SUB, 5, 160, string, TEXT_COLOR);
	}

#ifdef UCI_TEXT_OUT
   va_list arg_list;
/*   printf(format, arg_list);
   printf("\n");*/
   char string[1024]; //was 4096

   ASSERT(format!=NULL);

   va_start(arg_list,format);
   vsprintf(string,format,arg_list);
   va_end(arg_list);
   fprintf(stdout,"%s\n",string);
#endif
}

// string_equal()
static bool string_equal(const char s1[], const char s2[]) {

   ASSERT(s1!=NULL);
   ASSERT(s2!=NULL);

   return strcmp(s1,s2) == 0;
}
// end of protocol.cpp

